/*
 * db_creator.cpp
 *
 *  Created on: 2009/10/06
 *      Author: shu
 */
#include <ios>
#include <new>
#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <sstream>
#include <stdint.h>
#include <math.h>
#include <limits.h>
#include <string.h>
#include "logger.h"
#include "protein_type.h"
#include "encoder.h"
#include "sequence.h"
#include "sequence_reader.h"
#include "cached_sa_sequence_index.h"
#include "fmi_sequence_index.h"
#include "aligner_parameters.h"
#include "chunk.h"
#include "db.h"
#include "db_creator.h"

using namespace std;

void DbCreator::Run(const DbCreatorParameters &paramters)
{
  reader_ = CreateSequenceReader(paramters.GetInputName().c_str(), FASTA);
  vector<SequencePtr> sequences;
  uint32_t number_chunks = 0;
  uint32_t max_length = 0;
  uint64_t db_length = 0;
  uint64_t number_sequences = 0;
  Logger *logger = Logger::GetInstance();
  while (1) {
    ReadSequences(paramters.GetMaxChunkSequenceLength(), sequences);
    if (sequences.empty()) {
      break;
    }
    for (vector<SequencePtr>::iterator s = sequences.begin(); s != sequences.end(); ++s) {
      db_length += (*s)->GetSequenceData().length();
    }
    number_sequences += sequences.size();
    stringstream ss;
    ss.str("");
    ss << paramters.GetDbName() << "_" << number_chunks;
    string chunk_name = ss.str();
    Code *sequences_data;
    uint32_t *positions;
    uint32_t sequences_data_length;
    Encode(sequences, *(paramters.GetSequenceType()), &sequences_data,&sequences_data_length, &positions);
    if (max_length < sequences_data_length) {
      max_length = sequences_data_length;
    }
    SaveChunkInformation(chunk_name, sequences.size(), sequences_data_length);
    SaveNames(chunk_name, sequences);
    SaveSequences(chunk_name, sequences_data, sequences_data_length);
    SavePositions(chunk_name, positions, sequences.size());
    SaveIndex(chunk_name, *(paramters.GetSequenceType()), sequences_data, sequences_data_length, paramters.GetMaxCachedOffset());

    delete [] sequences_data;
    delete [] positions;
    ++number_chunks;
  }
  SaveDbInformation(paramters.GetDbName(), number_chunks, max_length, db_length, number_sequences);
  stringstream m;
  m << "The number of chunks :"<<number_chunks;
  logger->Log(m.str());
  m.str("");
  
  m << "Max length of a chunk : " << max_length;
  logger->Log(m.str());
  m.str("");
  m << "Total database length : " << db_length;
  logger->Log(m.str());
  m.str("");
  m << "Total number of sequences : " << number_sequences;
  logger->Log(m.str());
  m.str("");

  
}

void DbCreator::ReadSequences(uint32_t max_chunk_sequences_length, vector<SequencePtr> &sequences) {
  sequences.clear();
  uint32_t sum_length = 0;
  if (!next_sequence_) {
    next_sequence_ = reader_->Read();
  }

  if (!next_sequence_) {
    return;
  }

  sum_length = next_sequence_->GetSequenceData().length() + 2;
  if (sum_length >= max_chunk_sequences_length) {
    cerr << "error: the max chunk sequences length is too short." << endl;
    return;
  }

  sequences.push_back(next_sequence_);
  while ((next_sequence_ = reader_->Read())) {
    if (sum_length +  next_sequence_->GetSequenceData().length() + 1 >= max_chunk_sequences_length) {
      break;
    }
    sum_length += next_sequence_->GetSequenceData().length() + 1;
    sequences.push_back(next_sequence_);
  }
}

void DbCreator::Encode(vector<SequencePtr> &sequences, const SequenceType &sequence_type, Code **encoded_sequences, uint32_t *sequences_length, uint32_t **positions) {
  uint32_t length = 1;
  uint32_t *pos = new uint32_t[sequences.size() + 1];
  pos[0] = length;
  for (uint32_t i = 0; i < sequences.size(); ++i) {
    length += sequences[i]->GetSequenceData().length() + 1;
    pos[i + 1] = length;
  }

  Code *seq = new Code[length];
  uint32_t offset = 1;
  seq[0] = AlignerParameters::kSequenceEnd;
  Encoder encoder( sequence_type );
  for (vector<SequencePtr>::iterator i = sequences.begin(); i != sequences.end(); ++i) {
    encoder.Encode((*i)->GetSequenceData(), &seq[offset]);
    offset += (*i)->GetSequenceData().length();
    seq[offset] = AlignerParameters::kSequenceEnd;
    ++offset;
  }

  *encoded_sequences = seq;
  *sequences_length = length;
  *positions = pos;
}


void DbCreator::SaveChunkInformation(string chunk_name, uint32_t number_sequences,
    uint32_t sequences_data_length) {
  string filename = Chunk::GetInformationFileName(chunk_name);
  ofstream out;
  out.open(filename.c_str(), ios::binary);
  out.write((char *) &number_sequences, sizeof(number_sequences));
  out.write((char *) &sequences_data_length, sizeof(sequences_data_length));
  out.close();
}


void DbCreator::SaveNames(string chunk_name, vector<SequencePtr> &sequences) {
  ofstream out;
  string filename = Chunk::GetNamesFileName(chunk_name);
  out.open(filename.c_str());
  for (vector<SequencePtr>::iterator i = sequences.begin(); i != sequences.end(); ++i) {
    out << (*i)->GetName() << endl;
  }
  out.close();
}

void DbCreator::SaveSequences(string chunk_name, Code sequences_data[],
    uint32_t sequences_data_length) {
  ofstream out;
  string filename = Chunk::GetSequencesFileName(chunk_name);
  out.open(filename.c_str(), ios::binary);
  out.write((char *) sequences_data, sizeof(Code)*sequences_data_length);
  out.close();
}

void DbCreator::SavePositions(string chunk_name, uint32_t positions[],
    uint32_t number_sequences) {
  ofstream out;
  string filename = Chunk::GetOffsetsFileName(chunk_name);
  out.open(filename.c_str(), ios::binary);
  out.write((char *) positions, sizeof(uint32_t)*number_sequences);
  out.close();
}

void DbCreator::SaveIndex(string chunk_name, const SequenceType &sequence_type, Code sequences_data[], uint32_t sequences_data_length, uint32_t max_cached_offset)
{
  Encoder encoder( sequence_type );
  CachedSaSequenceIndex seq_index(sequences_data, sequences_data_length, max_cached_offset, encoder);
  string filename = Chunk::GetIndexFileName(chunk_name);
  ofstream out;
  out.open(filename.c_str(), ios::binary);
  seq_index.Save(out);
  out.close();
}

void DbCreator::SaveDbInformation(string db_name, uint32_t number_chunks, uint32_t max_length, uint64_t db_length, uint64_t number_sequences) {
  string filename = Db::GetInformationFileName(db_name);
  ofstream out;
  out.open(filename.c_str(), ios::binary);
  out.write((char *) &number_chunks, sizeof(number_chunks));
  out.write((char *) &max_length, sizeof(max_length));
  out.write((char *) &db_length, sizeof(db_length));
  out.write((char *) &number_sequences, sizeof(number_sequences));
  out.close();
}
