/*
 * index.cpp
 *
 *  Created on: 2009/10/19
 *      Author: shu
 */

#include <iostream>
#include <stdint.h>
#include <limits>
#include <vector>
#include <string.h>
#include <cstdlib>
#include "bit_vector.h"
#include "succinct_bit_vector.h"
#include "sais.h"
#include "fm_index.h"

using namespace std;

FmIndex::FmIndex()
:primary_(0), range_(0), number_alphabet_(0), log_interval_(0), length_(0), nodes_depth_list_(NULL),
 nodes_code_list_(NULL), codes_(NULL), decodes_(NULL), alphabet_offset_(NULL), children_(NULL),
 positions_(NULL), bit_vectors_(NULL), succint_bit_vectors_(NULL)
{}

FmIndex::FmIndex(uint8_t text[], uint32_t length, uint32_t log_interval)
:primary_(0), range_(0), number_alphabet_(0), log_interval_(0), length_(length), nodes_depth_list_(NULL),
 nodes_code_list_(NULL), codes_(NULL), decodes_(NULL), alphabet_offset_(NULL), children_(NULL),
 positions_(NULL), bit_vectors_(NULL), succint_bit_vectors_(NULL)
{
  Rebuild(text, length, log_interval);
}

FmIndex::FmIndex(ifstream &is)
:primary_(0), range_(0), number_alphabet_(0), log_interval_(0), length_(0), nodes_depth_list_(NULL),
 nodes_code_list_(NULL), codes_(NULL), decodes_(NULL), alphabet_offset_(NULL), children_(NULL),
 positions_(NULL), bit_vectors_(NULL), succint_bit_vectors_(NULL)
{
  Load(is);
}

void FmIndex::Rebuild(uint8_t *text, uint32_t length, uint32_t log_interval) {
  Delete();

  uint8_t max_alphabet = 0;
    for (uint32_t i = 0; i < length; ++i) {
      if (max_alphabet < text[i]) {
        max_alphabet = text[i];
      }
    }
    alphabet_offset_ = new uint32_t[max_alphabet + 2];
    memset(alphabet_offset_, 0, sizeof(uint32_t) * (max_alphabet + 2));
    for (uint32_t i = 0; i < length; ++i) {
      ++alphabet_offset_[text[i]];
    }

  #if 0
    {
      cerr << "count" << endl;
      for (uint8_t i = 0; i <= max_alphabet; ++i) {
        if (alphabet_offset_[i] > 0) {
          cerr << i << " " << alphabet_offset_[i] << endl;
        }
      }
    }
  #endif

    number_alphabet_ = 0;
    for (uint8_t i = 0; i <= max_alphabet; ++i) {
      if (alphabet_offset_[i] > 0) {
        ++number_alphabet_;
      }
    }

    if (number_alphabet_ < 2) {
      cerr << "error : number of alphabet is too small." << endl;
      throw -1;
    }

    log_interval_ = log_interval;
    int *sa = new int[length + 1];
    uint8_t *bwt = new uint8_t[length];
    memset(sa, 0, sizeof(int) * (length + 1));

    sais(text, sa, length);

    uint32_t interval_size = (1 << log_interval_);
    uint32_t position_mask = (1 << log_interval_) - 1;
    uint32_t positions_length = (length + interval_size) / interval_size;
    positions_ = new uint32_t[positions_length];
    positions_[0] = length;
    for (uint32_t i = 1; i <= length; ++i) {
      if (!(i & position_mask)) {
        positions_[i >> log_interval_] = sa[i - 1];
      }
    }
    primary_ = sais_bwt(text, bwt, sa, length);

    delete[] sa;

    range_ = static_cast<uint32_t> (max_alphabet) + 1;
    codes_ = new uint32_t[range_];
    children_ = new uint32_t[(number_alphabet_ - 1) * 2];
    bit_vectors_ = new BitVector*[number_alphabet_ - 1];

    for (uint32_t i = 0; i < number_alphabet_ - 1; ++i) {
      bit_vectors_[i] = NULL;
    }

    // build Huffman tree
    BuildTree();

  #if 0
    {
      cerr << "codes" << endl;
      for (uint8_t i = 0; i <= max_alphabet; ++i) {
        if (alphabet_offset_[i] > 0) {
          cerr << i << " " << codes_[i] << endl;
        }
      }
    }
  #endif

    // set decode
    SetDecodes();

    // bit set
    uint32_t* bit_vector_count = new uint32_t[number_alphabet_ - 1];
    memset(bit_vector_count, 0, sizeof(uint32_t) * (number_alphabet_ - 1));
    for (uint32_t i = 0; i < length; ++i) {
      Insert(codes_[bwt[i]], bit_vector_count);
    }

    delete[] bwt;
    delete[] bit_vector_count;

    // set succinct bit vector
    succint_bit_vectors_ = new SuccinctBitVector*[number_alphabet_ - 1];
    for (uint32_t i = 0; i < number_alphabet_ - 1; ++i) {
      if (bit_vectors_[i] != NULL) {
        succint_bit_vectors_[i] = new SuccinctBitVector(*bit_vectors_[i]);
      } else {
        succint_bit_vectors_[i] = NULL;
      }
    }

    uint32_t temp = 0;
    uint32_t sum = 0;
    for (uint8_t i = 0; i <= max_alphabet + 1; ++i) {
      temp = alphabet_offset_[i];
      alphabet_offset_[i] = sum;
      sum = sum + temp;
    }

  #if 0
    {
      cerr << "count" << endl;
      for (uint8_t i = 0; i <= max_alphabet + 1; ++i) {
        if (alphabet_offset_[i] > 0) {
          cerr << i << " " << alphabet_offset_[i] << endl;
        }
      }
    }
  #endif

    SetNodesList();
    length_ = length;
}

void FmIndex::SetDecodes() {
  uint32_t max_code = 0;
  for (uint32_t i = 0; i < range_; ++i) {
    if (codes_[i] != UINT_MAX && max_code < codes_[i]) {
      max_code = codes_[i];
    }
  }
  ++max_code;
  decodes_ = new uint8_t[max_code + 1];
  for (uint32_t i = 0; i <= max_code; ++i) {
    decodes_[i] = UCHAR_MAX;
  }

  for (uint32_t i = 0; i < range_; ++i) {
    if (codes_[i] != UINT_MAX) {
      decodes_[codes_[i]] = static_cast<uint8_t> (i);
    }
  }
#if 0
  for (uint32_t i = 0; i <= max_code; ++i) {
    cerr << static_cast<int>(decodes_[i]) << endl;
  }
#endif
}

void FmIndex::Insert(uint32_t code, uint32_t bit_vector_count[]) {
  uint32_t node = 0;
  uint32_t depth = 0;

  while (!IsLeaf(node)) {
    uint32_t bit = ((code >> depth) & 1);
    if (bit == 1) {
      bit_vectors_[node]->SetBit(bit_vector_count[node]);
      ++bit_vector_count[node];
      node = children_[node * 2 + 1];
    } else {
      ++bit_vector_count[node];
      node = children_[node * 2];
    }
    ++depth;
  }
}

void FmIndex::BuildTree() {
  uint32_t *parents = new uint32_t[(number_alphabet_ - 1) + number_alphabet_];
  vector<uint32_t> stack(number_alphabet_);
  vector<uint32_t> temp(number_alphabet_);

  memset(codes_, 0, sizeof(uint32_t) * range_);

  uint32_t temp_number_alphabet = 0;
  for (uint32_t i = 0; i < range_; ++i) {
    if (alphabet_offset_[i] > 0) {
      stack[temp_number_alphabet] = number_alphabet_ - 1 + temp_number_alphabet;
      temp[temp_number_alphabet] = alphabet_offset_[i];
      parents[stack[temp_number_alphabet]] = UINT_MAX;
      ++temp_number_alphabet;
    }
  }

  // insertion sort
  if (number_alphabet_ > 1) {
    for (uint32_t i = 1; i < number_alphabet_; ++i) {
      uint32_t t = stack[i];
      uint32_t u = temp[i];
      int j = i - 1;
      while ((j >= 0) && ((u > temp[j]) || ((u == temp[j]) && (t > stack[j])))) {
        stack[j + 1] = stack[j];
        temp[j + 1] = temp[j];
        --j;
      }
      stack[j + 1] = t;
      temp[j + 1] = u;
    }

    // choose two smallest element
    stack[number_alphabet_ - 1] = InitializeEdge(number_alphabet_ - 2,
        stack[number_alphabet_ - 1], stack[number_alphabet_ - 2], parents);
    temp[number_alphabet_ - 1] += temp[number_alphabet_ - 2];

    bit_vectors_[number_alphabet_ - 2] = new BitVector(temp[number_alphabet_
        - 1]);

    int l = number_alphabet_ - 1;
    int m = number_alphabet_ - 2;
    int r = number_alphabet_ - 3;

    while (m > 0) {
      uint32_t t1;
      uint32_t u;

      // left
      if ((r < 0) || temp[l] < temp[r]) {
        t1 = temp[l];
        u = stack[l];
        --l;
      } else {
        t1 = temp[r];
        u = stack[r];
        --r;
      }

      // right
      uint32_t t2;
      uint32_t v;
      if ((r < 0) || ((m < l) && (temp[l] < temp[r]))) {
        t2 = temp[l];
        v = stack[l];
        --l;
      } else {
        t2 = temp[r];
        v = stack[r];
        --r;
      }
      stack[m] = InitializeEdge(m - 1, u, v, parents);
      temp[m] = t1 + t2;

      bit_vectors_[m - 1] = new BitVector(temp[m]);
      --m;
    }
    stack[0] = stack[1];
  }

#if 0
  cerr << "parent" << endl;
  uint32_t s = (number_alphabet_ - 1) + number_alphabet_;
  for (uint32_t i = 0; i < s; ++i) {
    cerr << i << " " << parents[i] << endl;
  }

  cerr << "children" << endl;
  s = (number_alphabet_ - 1) * 2;
  for (uint32_t i = 0; i < s; ++i) {
    cerr << i << " " << children_[i] << endl;
  }
#endif
  // set codes
  temp_number_alphabet = 0;
  for (uint32_t i = 0; i < range_; ++i) {
    if (alphabet_offset_[i] > 0) {
      uint32_t code = 0;
      uint32_t u = number_alphabet_ - 1 + temp_number_alphabet;
      while (parents[u] != UINT_MAX) {
        uint32_t v = parents[u];
        code <<= 1;
        if (u == children_[2 * v + 1]) {
          code |= 1;
        }
        u = v;
      }
      codes_[i] = code;
      ++temp_number_alphabet;
    } else {
      codes_[i] = UINT_MAX;
    }
  }

  if (parents) {
    delete[] parents;
  }
}

void FmIndex::SetNodesList() {
  uint32_t list_length = number_alphabet_ - 1;
  nodes_depth_list_ = new uint32_t[list_length];
  nodes_code_list_ = new uint32_t[list_length];

  nodes_depth_list_[0] = 0;
  nodes_code_list_[0] = 0;

  for (uint32_t i = 0; i < list_length; ++i) {
    uint32_t node;
    uint32_t parent_code = nodes_code_list_[i];
    uint32_t depth = nodes_depth_list_[i];

    // bit 1
    node = children_[2 * i + 1];
    if (!IsLeaf(node)) {
      nodes_code_list_[node] = parent_code | (1 << depth);
      nodes_depth_list_[node] = depth + 1;
    }

    // bit 0
    node = children_[2 * i];
    if (!IsLeaf(node)) {
      nodes_code_list_[node] = parent_code;
      nodes_depth_list_[node] = depth + 1;
    }
  }
}

uint32_t FmIndex::Locate(uint32_t position) {
  uint32_t code;
  uint32_t node;
  uint32_t depth;
  uint32_t position_mask = (1 << log_interval_) - 1;
  uint32_t t = 0;

  while (position_mask & position) {
    if (position == primary_) {
      return t;
    } else if (position < primary_) {
      ++position;
    }
    code = 0;
    node = 0;
    depth = 0;
    for (;; ++depth) {
      uint32_t bit = bit_vectors_[node]->GetBit(position - 1);
      if (bit == 1) {
        position = succint_bit_vectors_[node]->Rank1(position);
        code = code | (1 << depth);
      } else {
        position = position - succint_bit_vectors_[node]->Rank1(position);
      }
      node = children_[2 * node + bit];
      if (position == 0 || IsLeaf(node)) {
        break;
      }
    }
    position = alphabet_offset_[decodes_[code]] + position;
    ++t;
  }
  return positions_[position >> log_interval_] + t;

}

void FmIndex::Occur(uint32_t position, uint32_t *occured_list, uint8_t start_character, uint8_t end_character) {
  uint32_t list_length = number_alphabet_ - 1;
  uint32_t *nodes_occured_list;
  nodes_occured_list = (uint32_t *)malloc(list_length * sizeof(uint32_t));
  if(nodes_occured_list == NULL){fprintf(stderr, "[ERROR] can not allocate momory: FmIndex::Occur\n");exit(1);}

  if (end_character > range_) {
    end_character = static_cast<uint8_t>(range_);
  }
  memset(occured_list, 0, sizeof(uint32_t) * (end_character - start_character + 1));

  if (position < primary_) {
    ++position;
  }

  nodes_occured_list[0] = position;

  for (uint32_t i = 0; i < list_length; ++i) {
    uint32_t node;
    uint32_t parent_code = nodes_code_list_[i];
    uint32_t parent_position = nodes_occured_list[i];
    uint32_t rank1 = 0;
    if (parent_position > 0) {
      rank1 = succint_bit_vectors_[i]->Rank1(parent_position);
    }

    // bit 1
    node = children_[2 * i + 1];
    if (IsLeaf(node)) {
      occured_list[decodes_[parent_code | (1 << nodes_depth_list_[i])]] = rank1;
    } else {
      nodes_occured_list[node] = rank1;
    }

    // bit 0
    node = children_[2 * i];
    if (IsLeaf(node)) {
      occured_list[decodes_[parent_code]] = parent_position - rank1;
    } else {
      nodes_occured_list[node] = parent_position - rank1;
    }
  }
}

void FmIndex::Load(ifstream &is) {
  Delete();

  is.read((char *) &primary_, sizeof(primary_));
  is.read((char *) &range_, sizeof(range_));
  is.read((char *) &number_alphabet_, sizeof(number_alphabet_));
  is.read((char *) &log_interval_, sizeof(log_interval_));
  codes_ = new uint32_t[range_];
  is.read((char *) codes_, sizeof(codes_[0]) * range_);
  alphabet_offset_ = new uint32_t[range_ + 1];
  is.read((char *) alphabet_offset_, sizeof(alphabet_offset_[0]) * (range_ + 1));
  children_ = new uint32_t[(number_alphabet_ - 1) * 2];
  is.read((char *) children_, sizeof(children_[0]) * (number_alphabet_ - 1) * 2);
  bit_vectors_ = new BitVector*[number_alphabet_ - 1];
  succint_bit_vectors_ = new SuccinctBitVector*[number_alphabet_ - 1];
  for (uint32_t i = 0; i < number_alphabet_ - 1; ++i) {
    uint32_t size;
    uint32_t blocks_size;
    uint32_t *blocks;
    uint32_t level_a_size;
    uint32_t *level_a;
    uint32_t level_b_size;
    uint8_t *level_b;
    is.read((char *) &size, sizeof(size));
    is.read((char *) &blocks_size, sizeof(blocks_size));
    blocks = new uint32_t[blocks_size];
    is.read((char *) blocks, sizeof(blocks[0]) * blocks_size);
    is.read((char *) &level_a_size, sizeof(level_a_size));
    level_a = new uint32_t[level_a_size];
    is.read((char *) level_a, sizeof(level_a[0]) * level_a_size);
    is.read((char *) &level_b_size, sizeof(level_b_size));
    level_b = new uint8_t[level_b_size];
    is.read((char *) level_b, sizeof(level_b[0]) * level_b_size);
    bit_vectors_[i] = new BitVector(blocks, size);
    succint_bit_vectors_[i] = new SuccinctBitVector(*bit_vectors_[i],
        level_a_size, level_b_size, level_a, level_b);
  }
  uint32_t interval_size = (1 << log_interval_);
  uint32_t positions_length = (bit_vectors_[0]->GetSize() + interval_size)
      / interval_size;
  positions_ = new uint32_t[positions_length];
  is.read((char *) positions_, sizeof(positions_) * positions_length);
  length_ = positions_[0];

  SetDecodes();
  SetNodesList();
}
void FmIndex::Save(ofstream &os) {
   os.write((const char*) &primary_, sizeof(primary_));
   os.write((const char*) &range_, sizeof(range_));
   os.write((const char*) &number_alphabet_, sizeof(number_alphabet_));
   os.write((const char*) &log_interval_, sizeof(log_interval_));
   os.write((const char*) codes_, sizeof(codes_[0])*range_);
   os.write((const char*) alphabet_offset_, sizeof(alphabet_offset_[0])*(range_ + 1));
   os.write((const char*) children_, sizeof(children_[0])*(number_alphabet_ - 1)*2);
   for (uint32_t i = 0; i < number_alphabet_ - 1; ++i) {
     uint32_t size = bit_vectors_[i]->GetSize();
     uint32_t block_size = bit_vectors_[i]->GetBlocksSize();
     uint32_t *blocks = bit_vectors_[i]->GetBlocks();
     uint32_t level_a_size = succint_bit_vectors_[i]->GetLevelASize();
     uint32_t *level_a = succint_bit_vectors_[i]->GetLevelA();
     uint32_t level_b_size = succint_bit_vectors_[i]->GetLevelBSize();
     uint8_t *level_b = succint_bit_vectors_[i]->GetLevelB();
     os.write((const char*) &size, sizeof(size));
     os.write((const char*) &block_size, sizeof(block_size));
     os.write((const char*) blocks, sizeof(blocks[0])*block_size);
     os.write((const char*) &level_a_size, sizeof(level_a_size));
     os.write((const char*) level_a, sizeof(level_a[0])*level_a_size);
     os.write((const char*) &level_b_size, sizeof(level_b_size));
     os.write((const char*) level_b, sizeof(level_b[0])*level_b_size);
   }

   uint32_t interval_size = (1 << log_interval_);
   uint32_t positions_length = (bit_vectors_[0]->GetSize() + interval_size) / interval_size;
   os.write((const char*) positions_, sizeof(positions_[0])*positions_length);
}
